from typing import Callable
import subprocess

import blstatus.config as config
import blstatus.inhibit as inhibit


def format_mebibyte(value):
    if value < 1024:
        return ''.join([str(round(value, 1)) + ' MiB'])
    else:
        return ''.join([str(round((value / 1024), 1)) + ' GiB'])


class Memory:
    _publish_status = None
    _sys_signal_text = ''
    _gpu_signal_text = ''
    _spacer = ''
    sys_text = ''
    gpu_text = ''

    def __init__(self, publish_status: Callable[[], None], sys_signal_text='', gpu_signal_text='', spacer=''):
        self._publish_status = publish_status
        self._sys_signal_text = sys_signal_text if config.enable_signal_text else ''  # statuscmd signal text
        self._gpu_signal_text = gpu_signal_text if config.enable_signal_text else ''
        self._spacer = spacer
        self._update_text()

    def _update_text(self):
        """Update system and GPU memory status text"""

        # System memory
        sys_mem_used = format_mebibyte(
            int(subprocess.run(['free', '-m'],
                               capture_output=True,
                               text=True).stdout.split('\n')[1].split()[2]))

        self.sys_text = '{}sys {}{}'.format(self._sys_signal_text, sys_mem_used, self._spacer)

        # GPU (NVIDIA) memory
        if config.memory_enable_gpu:
            try:
                gpu_mem_used = format_mebibyte(
                    int(subprocess.run(['nvidia-smi', '--query-gpu=memory.used', '--format=csv,noheader'],
                                       capture_output=True,
                                       text=True).stdout[:-5]))

                self.gpu_text = '{}gpu {}{}'.format(self._gpu_signal_text, gpu_mem_used, self._spacer)

            except FileNotFoundError:
                self.gpu_text = ''
            except ValueError:
                self.gpu_text = ''

    def update_and_publish(self):
        """Callback for the scheduler job"""

        # Don't update if going to sleep
        if inhibit.value:
            return

        self._update_text()
        self._publish_status()



import os
import configparser
import ast
import sys


# Text to use between components
spacer = ' | '

# Disable if not using statuscmd
enable_signal_text = False

# Format string for the date command
date_time_format = '+\"%Y-%m-%d %A %-I:%M %P\"'

# Interval in seconds between updating memory status
memory_interval = 2.

# Enable using nvidia-smi to get GPU memory usage
memory_enable_gpu = False

# Dictionary of audio sink/source name suffixes to abbreviations
volume_source_sink_abbreviations = {
    'analog-stereo': 'A',  # Internal speaker
    'hdmi-stereo': 'H',  # External (HDMI) speaker
    'a2dp_sink': 'B',  # Bluetooth
    'handsfree_head_unit': 'B',  # Bluetooth
}

# Print audio sink/source names with:
#   import pulsectl
#   pulse = pulsectl.Pulse()
#   print([sink.name for sink in pulse.sink_list()])
#   print([source.name for source in pulse.source_list()])

# Abbreviation to use if a sink/source name suffix isn't found
volume_source_sink_unknown_abbreviation = 'U'


# Remove first or last character(s) if they are quotes
def remove_quotes(s):

  if len(s) <= 1:
    return s

  if (s[0] == '\'' and s[-1] == '\'') or (s[0] == '\"' and s[-1] == '\"'):
    s = s[1:-1]

  return s


def load():

    config_file_path = os.path.expanduser('~/.config/blstatus/config.ini')

    # Retrun if there's no user config file
    if not os.path.exists(config_file_path):
        return

    # Load config file
    _config = configparser.ConfigParser()
    _config.read(config_file_path)

    if 'general' in _config:
        if 'spacer' in _config['general']:
            global spacer
            spacer = remove_quotes(_config['general']['spacer'])

        if 'enable_signal_text' in _config['general']:
            global enable_signal_text
            enable_signal_text = _config['general'].getboolean('enable_signal_text')

    if 'date_time' in _config:
        if 'format' in _config['date_time']:
            global date_time_format
            date_time_format = remove_quotes(_config['date_time']['format'])

    if 'memory' in _config:
        if 'interval' in _config['memory']:
            global memory_interval
            memory_interval = _config['memory'].getfloat('interval')

        if 'enable_gpu' in _config['memory']:
            global memory_enable_gpu
            memory_enable_gpu = _config['memory'].getboolean('enable_gpu')

    if 'volume' in _config:
        if 'source_sink_abbreviations' in _config['volume']:
            global volume_source_sink_abbreviations
            dict_string = _config['volume']['source_sink_abbreviations']

            try:
                volume_source_sink_abbreviations = ast.literal_eval(dict_string)
            except ValueError:
                print('Unable to create volume_source_sink_abbreviations dictionary from string:', file=sys.stderr)
                print('    ' + dict_string, file=sys.stderr)

        if 'source_sink_unknown_abbreviation' in _config['volume']:
            global volume_source_sink_unknown_abbreviation
            volume_source_sink_unknown_abbreviation = remove_quotes(_config['volume']['source_sink_unknown_abbreviation'])



#!/usr/bin/env python3

import threading
import asyncio
import asyncio_glib
import Xlib.display
from apscheduler.schedulers.background import BackgroundScheduler
from pydbus import SystemBus

import blstatus.config as config
import blstatus.inhibit as inhibit

from blstatus.network import Network
from blstatus.memory import Memory
from blstatus.volume import Volume
from blstatus.battery import Battery
from blstatus.date_time import DateTime

lock = threading.Lock()

display = Xlib.display.Display()
root = display.screen().root

network = None
memory = None
volume = None
battery = None
date_time = None

scheduler = BackgroundScheduler()
asyncio.set_event_loop_policy(asyncio_glib.GLibEventLoopPolicy())
loop = asyncio.new_event_loop()
system_bus = SystemBus()

login1_proxy = None


def prepare_for_sleep(sleeping):

    global battery, date_time, network, inhibit, scheduler

    if not sleeping:
        # Resuming
        battery.update_text()
        date_time.update_text()
        network.update_proxies_and_text()
        publish()
        inhibit.value = False
        scheduler.resume()
    else:
        # Ignore network signals when going to sleep
        inhibit.value = True
        scheduler.pause()


def publish():

    global lock, network, memory, volume, battery, date_time, root, display

    """Publish status text to root window name"""
    if lock.acquire(blocking=False):
        status_text = ' ' \
                      f'{network.text}' \
                      f'{memory.sys_text}' \
                      f'{memory.gpu_text}' \
                      f'{volume.sink_text}' \
                      f'{volume.source_text}' \
                      f'{battery.text}' \
                      f'{date_time.text}' \
                      ' '

        root.set_wm_name(status_text)
        display.sync()
        lock.release()


def main():

    global login1_proxy, config, network, memory, volume, battery, date_time, scheduler, loop

    login1_proxy = system_bus.get('org.freedesktop.login1')
    login1_proxy.PrepareForSleep.connect(prepare_for_sleep)

    # Load configuration from file if it exists
    config.load()

    # The signal text arguments should be empty if not using statuscmd
    network = Network(system_bus, publish, '\x01', '\x02', config.spacer)
    memory = Memory(publish, '\x03', '\x04', config.spacer)
    volume = Volume(loop, publish, '\x05', '\x06', config.spacer)
    battery = Battery(system_bus, publish, '\x07', config.spacer)
    date_time = DateTime(publish, '\x08')

    # Update memory every 2 seconds
    scheduler.add_job(memory.update_and_publish, 'interval', seconds=config.memory_interval)

    # Update date_time on every minute
    scheduler.add_job(date_time.update_and_publish, 'cron', second=0)

    publish()

    scheduler.start()

    loop.run_until_complete(volume.run())


if __name__ == '__main__':
    main()



from typing import Callable
import signal
import asyncio
from contextlib import suppress
import pulsectl_asyncio

import blstatus.config as config
import blstatus.inhibit as inhibit


def get_abbreviation(name: str) -> str:
    for key, value in config.volume_source_sink_abbreviations.items():
        if name.endswith(key):
            return value

    return config.volume_source_sink_unknown_abbreviation


class Volume:
    _loop = None
    _publish_status = None
    _sink_signal_text = ''
    _source_signal_text = ''
    _spacer = ''
    sink_text = '- / '
    source_text = '- | '

    def __init__(self, loop, publish_status: Callable[[], None], sink_signal_text='', source_signal_text='', spacer=''):
        self._loop = loop
        self._publish_status = publish_status
        self._sink_signal_text = sink_signal_text if config.enable_signal_text else ''  # statuscmd signal text
        self._source_signal_text = source_signal_text if config.enable_signal_text else ''
        self._spacer = spacer

    def _update_sink_text(self, sink):
        sink_abbreviation = get_abbreviation(sink.name)

        if sink.mute:
            self.sink_text = '{}{}: 0% / '.format(self._sink_signal_text, sink_abbreviation)
        else:
            self.sink_text = '{}{}: {:.0f}% / '.format(self._sink_signal_text, sink_abbreviation,
                                                       sink.volume.value_flat * 100.)

    def _update_source_text(self, source):
        source_abbreviation = get_abbreviation(source.name)

        if source.mute:
            self.source_text = '{}{}: 0%{}'.format(self._source_signal_text, source_abbreviation, self._spacer)
        else:
            self.source_text = '{}{}: {:.0f}%{}'.format(self._source_signal_text, source_abbreviation,
                                                        source.volume.value_flat * 100., self._spacer)

    async def _listen(self, pulse: pulsectl_asyncio.PulseAsync):
        async for event in pulse.subscribe_events('server', 'sink', 'source'):

            # Don't update if going to sleep
            if inhibit.value:
              continue

            # Ignore anything but change events
            if event.t != 'change':
                continue

            server_info = await pulse.server_info()

            # Update sink for server or sink event
            if event.facility != 'source':
                default_sink = await pulse.get_sink_by_name(server_info.default_sink_name)
                self._update_sink_text(default_sink)

            # Update source for server or source event
            if event.facility != 'sink':
                default_source = await pulse.get_source_by_name(server_info.default_source_name)
                self._update_source_text(default_source)

            self._publish_status()

    async def run(self):
        async with pulsectl_asyncio.PulseAsync('status-bar-volume') as pulse:
            # First update
            server_info = await pulse.server_info()

            default_sink = await pulse.get_sink_by_name(server_info.default_sink_name)
            self._update_sink_text(default_sink)

            default_source = await pulse.get_source_by_name(server_info.default_source_name)
            self._update_source_text(default_source)

            # Create listen task
            listen_task = asyncio.create_task(self._listen(pulse))

            # Handle signals
            for sig in (signal.SIGTERM, signal.SIGHUP, signal.SIGINT):
                self._loop.add_signal_handler(sig, listen_task.cancel)

            with suppress(asyncio.CancelledError):
                await listen_task



from typing import Callable
import subprocess
from pydbus.bus import Bus

import blstatus.config as config

UPOWER_TYPE_BATTERY = 2
UPOWER_STATE_CHARGING = 1
UPOWER_STATE_DISCHARGING = 2
UPOWER_STATE_FULLY_CHARGED = 4


def get_state_abbreviation(value):
    """Return a single character abbreviation for the battery state"""
    if value == UPOWER_STATE_CHARGING:
        return '+'
    elif value == UPOWER_STATE_DISCHARGING:
        return '-'
    elif value == UPOWER_STATE_FULLY_CHARGED:
        return ''
    else:
        return '?'


def format_time(total_seconds):
    if total_seconds == 0:
        return ''

    total_minutes = total_seconds // 60

    hours = total_minutes // 60
    minutes = total_minutes % 60

    return ' ({}:{:0>2})'.format(hours, minutes)


class Battery:
    _publish_status = None
    _signal_text = ''
    _spacer = ''
    _system_bus = None
    _device_proxy = None
    text = ''

    def __init__(self, system_bus: Bus, publish_status: Callable[[], None], signal_text='', spacer=''):

        self._publish_status = publish_status  # Function to update the full status text
        self._signal_text = signal_text if config.enable_signal_text else ''  # statuscmd signal text
        self._spacer = spacer  # Spacer text added after status
        self._system_bus = system_bus

        upower_proxy = self._system_bus.get('org.freedesktop.UPower')
        devices = upower_proxy.EnumerateDevices()

        # For each UPower device
        for device in devices:
            device_proxy = self._system_bus.get('org.freedesktop.UPower', device)

            # Find the first battery
            if device_proxy.Type == UPOWER_TYPE_BATTERY:
                self._device_proxy = device_proxy
                break

        if self._device_proxy is not None:
            self._device_proxy.PropertiesChanged.connect(self._update_and_publish)
            self.update_text()  # Initial update of battery status

    def update_text(self):
        """Update battery status text"""

        # Get battery state
        state = self._device_proxy.State

        # Get the time string
        time = ''

        if state == UPOWER_STATE_CHARGING:
            time = format_time(self._device_proxy.TimeToFull)
        elif state == UPOWER_STATE_DISCHARGING:
            time = format_time(self._device_proxy.TimeToEmpty)

        if time == '':
            state = UPOWER_STATE_FULLY_CHARGED

        # Get battery charge percent
        battery_percent = round(self._device_proxy.Percentage)

        # Format status text
        self.text = f'{self._signal_text}bat {get_state_abbreviation(state)}{str(battery_percent)}%{time}{self._spacer}'

    def _update_and_publish(self, *params):
        """Callback for the PropertiesChanged signal of the battery device proxy"""
        self.update_text()  # Update battery status text
        self._publish_status()  # Publish full status text



from typing import Callable
import subprocess

import blstatus.config as config
import blstatus.inhibit as inhibit


class DateTime:
    _publish_status = None
    _signal_text = ''
    _spacer = ''
    text = ''

    def __init__(self, publish_status: Callable[[], None], signal_text='', spacer=''):
        self._publish_status = publish_status
        self._signal_text = signal_text if config.enable_signal_text else ''  # statuscmd signal text
        self._spacer = spacer
        self.update_text()

    def update_text(self):
        """Update the date/time status text"""

        # Run the date command to get the date/time text
        self.text = '{}{}{}'.format(self._signal_text,
                                    subprocess.run(['date', config.date_time_format],
                                                   capture_output=True,
                                                   text=True).stdout[1:-2], self._spacer)

    def update_and_publish(self):
        """Callback for the scheduler job"""

        # Don't update if going to sleep
        if inhibit.value:
            return

        self.update_text()
        self._publish_status()



from time import sleep
from typing import Callable
from pydbus.bus import Bus

import blstatus.config as config
import blstatus.inhibit as inhibit

NM_DEVICE_TYPE_ETHERNET = 1
NM_DEVICE_TYPE_WIFI = 2

NM_DEVICE_STATE_ACTIVATED = 100

active_connection_types = ['802-3-ethernet', '802-11-wireless']


def format_bps(value):
    if value >= 1000:
        return ''.join([str(int(value / 1000)) + ' Gbit/s'])
    else:
        return ''.join([str(value) + ' Mbit/s'])


def format_ssid(value_list):
    char_list = []
    for v in value_list:
        char_list.append(chr(v))

    return ''.join(char_list)


class Network:
    _publish_status = None
    _wifi_signal_text = ''
    _ethernet_signal_text = ''
    _spacer = ''
    text = ''
    _system_bus = None
    _nm_proxy = None
    _active_connection_proxies = []
    _active_connection_property_subs = []
    _device_proxies = []

    def __init__(self,
                 system_bus: Bus,
                 publish_status: Callable[[], None],
                 wifi_signal_text='',
                 ethernet_signal_text='',
                 spacer=''):

        self._publish_status = publish_status
        self._wifi_signal_text = wifi_signal_text if config.enable_signal_text else ''  # statuscmd signal text
        self._ethernet_signal_text = ethernet_signal_text if config.enable_signal_text else ''
        self._spacer = spacer
        self._system_bus = system_bus

        self._nm_proxy = self._system_bus.get('org.freedesktop.NetworkManager')

        self.update_proxies_and_text()
        self._nm_proxy.PropertiesChanged.connect(self._callback_nm)

    def update_proxies_and_text(self):
        """ Update active connection proxies, device proxies, and status text"""

        self._device_proxies.clear()

        # Update active connection proxies
        for active_connection in self._nm_proxy.ActiveConnections:

            try:
                active_connection_proxy = self._system_bus.get('org.freedesktop.NetworkManager', active_connection)
            except KeyError:
                continue

            global active_connection_types

            if active_connection_proxy.Type in active_connection_types:

                self._active_connection_property_subs.append(
                    active_connection_proxy.PropertiesChanged.connect(self._callback_active_connection))
                self._active_connection_proxies.append(active_connection_proxy)

                for device in active_connection_proxy.Devices:
                    device_proxy = self._system_bus.get('org.freedesktop.NetworkManager', device)
                    self._device_proxies.append(device_proxy)

        # Then update device proxies and status text
        self._update_device_proxies_and_text()

    def _update_device_proxies_and_text(self):
        """Update device proxies and status text"""

        self.text = ''

        for device_proxy in self._device_proxies:

            if device_proxy.State != NM_DEVICE_STATE_ACTIVATED:
                continue

            ip4_config_proxy = self._system_bus.get('org.freedesktop.NetworkManager', device_proxy.Ip4Config)

            # Wait until the address is valid
            while (len(ip4_config_proxy.AddressData) == 0):
              sleep(0.1)

            address = ip4_config_proxy.AddressData[0]["address"]

            if device_proxy.DeviceType == NM_DEVICE_TYPE_ETHERNET:
                if device_proxy.Speed == 0:
                    text = '{}{}: {}{}'.format(self._ethernet_signal_text,
                                               device_proxy.Interface,
                                               address,
                                               self._spacer)
                    self.text += text

                else:
                    text = '{}{}: {} ({}){}'.format(self._ethernet_signal_text, device_proxy.Interface,
                                                    address,
                                                    format_bps(device_proxy.Speed),
                                                    self._spacer)
                    self.text += text

            elif device_proxy.DeviceType == NM_DEVICE_TYPE_WIFI:
                active_access_point_proxy = self._system_bus.get('org.freedesktop.NetworkManager',
                                                                 device_proxy.ActiveAccessPoint)

                text = '{}{}: {} ({}){}'.format(self._wifi_signal_text, device_proxy.Interface,
                                                address,
                                                format_ssid(active_access_point_proxy.Ssid),
                                                self._spacer)
                self.text += text

    def _callback_nm(self, *params):
        """Callback for the PropertiesChanged signal of the NetworkManager proxy"""

        # Don't update if going to sleep
        if inhibit.value:
            return

        # If active connections have changed
        if 'ActiveConnections' in params[1]:
            # Update active connection proxies, device proxies, and status text
            self.update_proxies_and_text()
            self._publish_status()

    def _callback_active_connection(self, *params):
        """Callback for the PropertiesChanged signal of the active connection proxies"""

        # Don't update if going to sleep
        if inhibit.value:
            return

        # If the IPv4 address has changed
        if 'Ip4Config' in params[1] and params[1]['Ip4Config'] != '/':
            # Update device proxies and status text
            self._update_device_proxies_and_text()
            self._publish_status()
